from manifest_edit.plugin.mpd import ManifestIteratorPlugin
from manifest_edit.context import Context
from manifest_edit import libfmp4
from schema import Schema, Optional, Or
from manifest_edit.plugins.utility import renderFormatString
import re


class Plugin(ManifestIteratorPlugin):
    """
    Segment URL modifier plugin

    This plugin allows to modify the 'initialization' and 'media' attributes of
    the SegmentTemplate element (see 5.3.9). This is an example of such an
    element:

    .. code-block:: xml

      <SegmentTemplate
        timescale="48000"
        initialization="audio-$RepresentationID$.dash"
        media="audio-$RepresentationID$-$Time$.dash">
        <SegmentTimeline>
          <S t="0" d="125952" />
          <S d="128000" r="272" />
          <S d="158992" />
        </SegmentTimeline>
      </SegmentTemplate>

    It also allows to modify the baseURL element wherever it may be
    present in the manifest.

    In particular scenarios where users may want to serve segments from
    different servers or referring to individual ism server manifests,
    editing these attributes is essential to form the desired segment URLS.
    """

    _name = __name__

    _keys = ["segmentTemplate", "baseURL"]

    # This is just the specific ["how" config] for this plugin
    def schema(self):
        return Schema(
            {
                Optional(self._keys[0], default={}): {
                    Optional("initialization", default={}): Or(
                        {},
                        {
                            "match": lambda s: re.compile(s),
                            "replace": str,
                        },
                    ),
                    Optional("media", default={}): Or(
                        {},
                        {
                            "match": lambda s: re.compile(s),
                            "replace": str,
                        },
                    ),
                },
                Optional(self._keys[1], default={}): Or(
                    {},
                    {
                        "match": lambda s: re.compile(s),
                        "replace": str,
                    },
                ),
            }
        )

    def _attemptMod(self, value: str, match: str, replace: str) -> str:
        # We know that "match" is a proper regular expression
        # We don't know if the replace string is. If it is not, re.sub will
        # just return the original string

        (new_string, number_of_subs_made) = re.subn(match, replace, value)

        if number_of_subs_made == 0:
            # This is actually OK, it is possible that a user has selected
            # all representations (i.e. - id: '.*') but then chose a match
            # string that only matches some representations.
            Context.log_trace(
                f"When trying to modify string '{value}' "
                f"using match string '{match}' "
                f"and replace string '{replace}' "
                " no substitutions have been made!"
            )

        return new_string

    def _modifySegmentTemplate(self, segment_config, element):
        for attribute_name in ["initialization", "media"]:
            if segment_config[attribute_name]:
                # segmentTemplate is an optional attribute and as such
                # you must check if it's there. If it's not, just skip...
                if getattr(element, "segmentTemplate", None):
                    rendered_replace_string = renderFormatString(
                        element, segment_config[attribute_name]["replace"]
                    )
                    # Weird stuff can happen with supposedly "raw" string
                    # coming from pyyaml and unicode escapes, so prepare to
                    # debug
                    Context.log_trace(
                        f"Rendered string '{segment_config[attribute_name]['replace']}' into '{rendered_replace_string}'"
                    )
                    setattr(
                        element.segmentTemplate,
                        attribute_name,
                        self._attemptMod(
                            getattr(element.segmentTemplate, attribute_name),
                            segment_config[attribute_name]["match"],
                            rendered_replace_string,
                        ),
                    )
                else:
                    # We will not allow to *create* a segment template from
                    # scratch in manifest edit because of course this will
                    # never work. 
                    Context.log_trace(
                        f'No segmentTemplate element detected in this '
                        f'{type(element)}. Is the configuration correct?'
                    )

    def _modifyBaseURL(self, baseURL_config, element):
        if baseURL_config:
            rendered_replace_string = renderFormatString(
                element, baseURL_config["replace"]
            )
            Context.log_trace(
                f"Rendered string '{baseURL_config['replace']}' into '{rendered_replace_string}'"
            )
            # baseURL needs to be converted back and forth to string
            setattr(
                element,
                "baseURL",
                libfmp4.Url(
                    self._attemptMod(
                        getattr(element, "baseURL").__repr__(),
                        baseURL_config["match"],
                        rendered_replace_string,
                    )
                ),
            )

    def _modifyBaseURLOrSegmentTemplate(self, manifest, storage):
        for segment_config, element in self.config(manifest, storage):
            if segment_config.get(self._keys[1], None):
                self._modifyBaseURL(segment_config[self._keys[1]], element)
            elif segment_config.get(self._keys[0], None):
                self._modifySegmentTemplate(segment_config[self._keys[0]], element)

    def process(self, manifest, storage):
        self._modifyBaseURLOrSegmentTemplate(manifest, storage)
